@@ -1,47 +1,58 @@
+import base64
 import os
 import threading
+from typing import TYPE_CHECKING
 
-from PyQt5.QtCore import Qt, QAbstractListModel, QModelIndex
-from PyQt5.QtCore import pyqtProperty, pyqtSignal, pyqtSlot, QObject
+from PyQt6.QtCore import Qt, QAbstractListModel, QModelIndex
+from PyQt6.QtCore import pyqtProperty, pyqtSignal, pyqtSlot, QObject
 
 from electrum.i18n import _
 from electrum.logging import get_logger
-from electrum.util import WalletFileException, standardize_path
-from electrum.wallet import Abstract_Wallet
+from electrum.util import WalletFileException, standardize_path, InvalidPassword, send_exception_to_crash_reporter
 from electrum.plugin import run_hook
 from electrum.lnchannel import ChannelState
-from electrum.daemon import Daemon
+from electrum.bitcoin import is_address
+from electrum.bitcoin import verify_usermessage_with_address
+from electrum.storage import StorageReadWriteError
 
 from .auth import AuthMixin, auth_protect
 from .qefx import QEFX
 from .qewallet import QEWallet
-from .qewalletdb import QEWalletDB
-from .qewizard import QENewWalletWizard, QEServerConnectWizard
+from .qewizard import QENewWalletWizard, QEServerConnectWizard, QETermsOfUseWizard
+
+if TYPE_CHECKING:
+    from electrum.daemon import Daemon
+    from electrum.plugin import Plugins
+
 
 # wallet list model. supports both wallet basenames (wallet file basenames)
 # and whole Wallet instances (loaded wallets)
+from .util import check_password_strength
+
+
 class QEWalletListModel(QAbstractListModel):
     _logger = get_logger(__name__)
 
     # define listmodel rolemap
-    _ROLE_NAMES= ('name','path','active')
-    _ROLE_KEYS = range(Qt.UserRole, Qt.UserRole + len(_ROLE_NAMES))
+    _ROLE_NAMES= ('name', 'path', 'active')
+    _ROLE_KEYS = range(Qt.ItemDataRole.UserRole, Qt.ItemDataRole.UserRole + len(_ROLE_NAMES))
     _ROLE_MAP  = dict(zip(_ROLE_KEYS, [bytearray(x.encode()) for x in _ROLE_NAMES]))
 
     def __init__(self, daemon, parent=None):
         QAbstractListModel.__init__(self, parent)
         self.daemon = daemon
+        self._wallets = []
         self.reload()
 
     def rowCount(self, index):
-        return len(self.wallets)
+        return len(self._wallets)
 
     def roleNames(self):
         return self._ROLE_MAP
 
     def data(self, index, role):
-        (wallet_name, wallet_path) = self.wallets[index.row()]
-        role_index = role - Qt.UserRole
+        (wallet_name, wallet_path) = self._wallets[index.row()]
+        role_index = role - Qt.ItemDataRole.UserRole
         role_name = self._ROLE_NAMES[role_index]
         if role_name == 'name':
             return wallet_name
@@ -54,7 +65,7 @@ class QEWalletListModel(QAbstractListModel):
     def reload(self):
         self._logger.debug('enumerating available wallets')
         self.beginResetModel()
-        self.wallets = []
+        self._wallets = []
         self.endResetModel()
 
         available = []
@@ -65,21 +76,21 @@ class QEWalletListModel(QAbstractListModel):
                     available.append(i.path)
         for path in sorted(available):
             wallet = self.daemon.get_wallet(path)
-            self.add_wallet(wallet_path = path)
+            self.add_wallet(wallet_path=path)
 
     def add_wallet(self, wallet_path):
-        self.beginInsertRows(QModelIndex(), len(self.wallets), len(self.wallets))
+        self.beginInsertRows(QModelIndex(), len(self._wallets), len(self._wallets))
         wallet_name = os.path.basename(wallet_path)
         wallet_path = standardize_path(wallet_path)
         item = (wallet_name, wallet_path)
-        self.wallets.append(item)
+        self._wallets.append(item)
         self.endInsertRows()
 
     def remove_wallet(self, path):
         i = 0
         wallets = []
         remove = -1
-        for wallet_name, wallet_path in self.wallets:
+        for wallet_name, wallet_path in self._wallets:
             if wallet_path == path:
                 remove = i
             else:
@@ -88,12 +99,12 @@ class QEWalletListModel(QAbstractListModel):
 
         if remove >= 0:
             self.beginRemoveRows(QModelIndex(), i, i)
-            self.wallets = wallets
+            self._wallets = wallets
             self.endRemoveRows()
 
     @pyqtSlot(str, result=bool)
     def wallet_name_exists(self, name):
-        for wallet_name, wallet_path in self.wallets:
+        for wallet_name, wallet_path in self._wallets:
             if name == wallet_name:
                 return True
         return False
@@ -101,19 +112,23 @@ class QEWalletListModel(QAbstractListModel):
     @pyqtSlot(str)
     def updateWallet(self, path):
         i = 0
-        for wallet_name, wallet_path in self.wallets:
+        for wallet_name, wallet_path in self._wallets:
             if wallet_path == path:
                 mi = self.createIndex(i, i)
                 self.dataChanged.emit(mi, mi, self._ROLE_KEYS)
                 return
             i += 1
 
+
 class QEDaemon(AuthMixin, QObject):
+    instance = None  # type: Optional[QEDaemon]
+
     _logger = get_logger(__name__)
 
     _available_wallets = None
     _current_wallet = None
     _new_wallet_wizard = None
+    _terms_of_use_wizard = None
     _server_connect_wizard = None
     _path = None
     _name = None
@@ -126,49 +141,52 @@ class QEDaemon(AuthMixin, QObject):
     availableWalletsChanged = pyqtSignal()
     fxChanged = pyqtSignal()
     newWalletWizardChanged = pyqtSignal()
+    termsOfUseWizardChanged = pyqtSignal()
     serverConnectWizardChanged = pyqtSignal()
     loadingChanged = pyqtSignal()
-    passwordChangeFailed = pyqtSignal()
     requestNewPassword = pyqtSignal()
 
-    walletLoaded = pyqtSignal([str,str], arguments=['name','path'])
-    walletRequiresPassword = pyqtSignal([str,str], arguments=['name','path'])
+    walletLoaded = pyqtSignal([str, str], arguments=['name', 'path'])
+    walletRequiresPassword = pyqtSignal([str, str], arguments=['name', 'path'])
     walletOpenError = pyqtSignal([str], arguments=["error"])
-    walletDeleteError = pyqtSignal([str,str], arguments=['code', 'message'])
+    walletDeleteError = pyqtSignal([str, str], arguments=['code', 'message'])
 
-    def __init__(self, daemon: 'Daemon', parent=None):
+    def __init__(self, daemon: 'Daemon', plugins: 'Plugins', parent=None):
         super().__init__(parent)
+        if QEDaemon.instance:
+            raise RuntimeError('There should only be one QEDaemon instance')
+        QEDaemon.instance = self
         self.daemon = daemon
+        self.plugins = plugins
         self.qefx = QEFX(daemon.fx, daemon.config)
 
         self._backendWalletLoaded.connect(self._on_backend_wallet_loaded)
 
-        self._walletdb = QEWalletDB()
-        self._walletdb.validPasswordChanged.connect(self.passwordValidityCheck)
-        self._walletdb.walletOpenProblem.connect(self.onWalletOpenProblem)
-
     @pyqtSlot()
     def passwordValidityCheck(self):
         if not self._walletdb._validPassword:
             self.walletRequiresPassword.emit(self._name, self._path)
 
-    @pyqtSlot(str)
-    def onWalletOpenProblem(self, error):
-        self.walletOpenError.emit(error)
-
     @pyqtSlot()
     @pyqtSlot(str)
     @pyqtSlot(str, str)
     def loadWallet(self, path=None, password=None):
+        if self._loading:
+            return
+        self._loading = True
+
         if path is None:
-            self._path = self.daemon.config.get('wallet_path') # command line -w option
+            self._path = self.daemon.config.get('wallet_path')  # command line -w option
             if self._path is None:
-                self._path = self.daemon.config.get('gui_last_wallet')
+                self._path = self.daemon.config.CURRENT_WALLET
         else:
             self._path = path
         if self._path is None:
+            self._loading = False
             return
 
+        self.loadingChanged.emit()
+
         self._path = standardize_path(self._path)
         self._name = os.path.basename(self._path)
 
@@ -181,35 +199,37 @@ class QEDaemon(AuthMixin, QObject):
         if not password:
             password = self._password
 
-        wallet_already_open = self._path in self.daemon._wallets
-
-        if not wallet_already_open:
-            # pre-checks, let walletdb trigger any necessary user interactions
-            self._walletdb.path = self._path
-            self._walletdb.password = password
-            self._walletdb.verify()
-            if not self._walletdb.ready:
-                return
+        wallet_already_open = self.daemon.get_wallet(self._path)
+        if wallet_already_open is not None:
+            wallet_already_open_password = QEWallet.getInstanceFor(wallet_already_open).password
 
         def load_wallet_task():
-            self._loading = True
-            self.loadingChanged.emit()
-
+            success = False
             try:
-                local_password = password # need this in local scope
-                wallet = self.daemon.load_wallet(self._path, local_password)
+                local_password = password  # need this in local scope
+                wallet = None
+                try:
+                    wallet = self.daemon.load_wallet(self._path, local_password, upgrade=True)
+                except InvalidPassword:
+                    self.walletRequiresPassword.emit(self._name, self._path)
+                except FileNotFoundError:
+                    self.walletOpenError.emit(_('File not found'))
+                except StorageReadWriteError:
+                    self.walletOpenError.emit(_('Could not read/write file'))
+                except WalletFileException as e:
+                    self.walletOpenError.emit(_('Could not open wallet: {}').format(str(e)))
+                    if e.should_report_crash:
+                        send_exception_to_crash_reporter(e)
 
                 if wallet is None:
-                    self._logger.info('could not open wallet')
-                    self.walletOpenError.emit('could not open wallet')
                     return
 
-                if wallet_already_open:
+                if wallet_already_open is not None:
                     # wallet already open. daemon.load_wallet doesn't mind, but
                     # we need the correct current wallet password below
-                    local_password = QEWallet.getInstanceFor(wallet).password
+                    local_password = wallet_already_open_password
 
-                if self.daemon.config.get('single_password'):
+                if self.daemon.config.WALLET_USE_SINGLE_PASSWORD:
                     self._use_single_password = self.daemon.update_password_for_directory(old_password=local_password, new_password=local_password)
                     self._password = local_password
                     self.singlePasswordChanged.emit()
@@ -217,31 +237,31 @@ class QEDaemon(AuthMixin, QObject):
                 else:
                     self._logger.info('use single password disabled by config')
 
-                self.daemon.config.save_last_wallet(wallet)
-
                 run_hook('load_wallet', wallet)
 
+                success = True
                 self._backendWalletLoaded.emit(local_password)
-            except WalletFileException as e:
-                self._logger.error(str(e))
-                self.walletOpenError.emit(str(e))
             finally:
-                self._loading = False
-                self.loadingChanged.emit()
+                if not success:  # if successful, _loading guard will be reset by _on_backend_wallet_loaded
+                    self._loading = False
+                    self.loadingChanged.emit()
 
-        threading.Thread(target=load_wallet_task, daemon=True).start()
+        threading.Thread(target=load_wallet_task, daemon=False).start()
 
     @pyqtSlot()
     @pyqtSlot(str)
-    def _on_backend_wallet_loaded(self, password = None):
+    def _on_backend_wallet_loaded(self, password=None):
         self._logger.debug('_on_backend_wallet_loaded')
-        wallet = self.daemon._wallets[self._path]
+        wallet = self.daemon.get_wallet(self._path)
+        assert wallet is not None
         self._current_wallet = QEWallet.getInstanceFor(wallet)
         self.availableWallets.updateWallet(self._path)
-        self._current_wallet.password = password if password else None
+        wallet.unlock(password or None)  # not conditional on wallet.requires_unlock in qml, as
+        # the auth wrapper doesn't pass the entered password, but instead we rely on the password in memory
+        self._loading = False
+        self.loadingChanged.emit()
         self.walletLoaded.emit(self._name, self._path)
 
-
     @pyqtSlot(QEWallet)
     @pyqtSlot(QEWallet, bool)
     @pyqtSlot(QEWallet, bool, bool)
@@ -322,19 +342,19 @@ class QEDaemon(AuthMixin, QObject):
         else:
             self.currentWallet.requestNewPassword.emit()
 
-    @pyqtSlot(str)
+    @pyqtSlot(str, result=bool)
     def setPassword(self, password):
         assert self._use_single_password
         assert password
         if not self.daemon.update_password_for_directory(old_password=self._password, new_password=password):
-            self.passwordChangeFailed.emit()
-            return
+            return False
         self._password = password
+        return True
 
     @pyqtProperty(QENewWalletWizard, notify=newWalletWizardChanged)
     def newWalletWizard(self):
         if not self._new_wallet_wizard:
-            self._new_wallet_wizard = QENewWalletWizard(self)
+            self._new_wallet_wizard = QENewWalletWizard(self, self.plugins)
 
         return self._new_wallet_wizard
 
@@ -345,6 +365,32 @@ class QEDaemon(AuthMixin, QObject):
 
         return self._server_connect_wizard
 
+    @pyqtProperty(QETermsOfUseWizard, notify=termsOfUseWizardChanged)
+    def termsOfUseWizard(self):
+        if not self._terms_of_use_wizard:
+            self._terms_of_use_wizard = QETermsOfUseWizard(self)
+        return self._terms_of_use_wizard
+
     @pyqtSlot()
     def startNetwork(self):
         self.daemon.start_network()
+
+    @pyqtSlot(str, str, str, result=bool)
+    def verifyMessage(self, address, message, signature):
+        address = address.strip()
+        message = message.strip().encode('utf-8')
+        if not is_address(address):
+            return False
+        try:
+            # This can throw on invalid base64
+            sig = base64.b64decode(str(signature.strip()), validate=True)
+            verified = verify_usermessage_with_address(address, sig, message)
+        except Exception as e:
+            verified = False
+        return verified
+
+    @pyqtSlot(str, result=int)
+    def passwordStrength(self, password):
+        if len(password) == 0:
+            return 0
+        return check_password_strength(password)[0]
